@@ -1,7 +1,6 @@
-import pickle
+import dill
 import numpy as np
 import pandas as pd
-from copy import deepcopy
 from sklearn.cluster import KMeans
 from zedsuite.zutil import Llk, Lsmash, Prun, DrawPFSA
 from zedsuite.genesess import GenESeSS
@@ -85,7 +84,7 @@ class AnomalyDetection:
         self.__calculate_cluster_labels()
         self.__calculate_cluster_PFSAs()
         self.__calculate_PFSA_stats()
-        # self.__reduce_clusters()
+        self.__reduce_clusters()
         self.fitted = True
 
         return self
@@ -110,7 +109,7 @@ class AnomalyDetection:
         num_predictions = 0
         # commonly want to find anomalies in original data
         if X is None:
-            # occurs when model is loaded from pickle file
+            # occurs when model is loaded from file
             if self.quantized_data is None:
                 raise ValueError("Original data not found. Pass data to predict().")
             data = self.quantized_data.drop(columns=["cluster"], axis=1)
@@ -119,23 +118,17 @@ class AnomalyDetection:
             data = self.__quantize(X)
             num_predictions = 1 if type(X) is pd.Series else data.shape[0]
 
-        cluster_llks = np.empty([self.n_clusters, num_predictions], dtype=np.float64)
-        anomaly_vec = np.zeros(num_predictions, dtype=np.int64)
+        cluster_llks = np.empty(shape=(self.n_clusters, num_predictions), dtype=np.float32)
         for i in range(self.n_clusters):
-            curr_llks = Llk(data=data, pfsafile=self.cluster_PFSA_files[i]).run()
-            cluster_llks[i] = curr_llks
-            # classify llk as anomaly if greater than X standard deviations above the mean
-            upper_bound = self.PFSA_llk_means[i] + (self.PFSA_llk_stds[i] * self.anomaly_sensitivity)
-            # diffs = curr_llks - upper_bound
-            for j, llk in enumerate(curr_llks):
-                anomaly_vec[j] += 1 if llk > upper_bound else 0
+            cluster_llks[i] = np.asarray(Llk(data=data, pfsafile=self.cluster_PFSA_files[i]).run(), dtype=np.float32)
+
+        # consider to be anomaly if all llks above specified upper bound (X standard deviations above the mean)
+        upper_bounds = self.PFSA_llk_means + (self.PFSA_llk_stds * self.anomaly_sensitivity)
+        predictions = np.all(cluster_llks.T > upper_bounds, axis=1)
 
         self.cluster_llks = cluster_llks
         self.closest_match = np.argmin(cluster_llks, axis=0)
 
-        # consider to be anomaly if all llks above specified upper bound
-        predictions = [x == self.n_clusters for x in anomaly_vec]
-
         if len(predictions) == 1:
             predictions = predictions[0]
 
@@ -152,13 +145,29 @@ class AnomalyDetection:
         if not self.fitted:
             raise ValueError("Model has not been fit yet")
 
-        model = deepcopy(self)
-        model.quantized_data = None
-        model.cluster_PFSA_files = []
-        model.cluster_PFSA_pngs = []
+        metadata = {
+            "modeltype": type(self),
+            "user_params": {
+                "anomaly_sensitivity": self.anomaly_sensitivity,
+                "n_clusters": self.n_clusters,
+                "clustering_alg": self.clustering_alg,
+                "quantize": self.quantize,
+                "quantize_type": self.quantize_type,
+                "eps": self.eps,
+                "verbose": self.verbose,
+            },
+            "fitted_params": {
+                "quantizer_parameters": None if not self.quantize else self.quantizer.parameters ,
+                "quantizer_feature_order": None if not self.quantize else self.quantizer._feature_order,
+                "cluster_labels": self.cluster_labels,
+                "cluster_PFSA_info": self.cluster_PFSA_info,
+                "PFSA_llk_means": self.PFSA_llk_means,
+                "PFSA_llk_stds": self.PFSA_llk_stds
+            }
+        }
 
         with open(path, "wb") as f:
-            pickle.dump(model, f)
+            dill.dump(metadata, f)
 
 
     @staticmethod
@@ -169,25 +178,33 @@ class AnomalyDetection:
                 path (str): path to saved model
 
             Returns:
-                AnomalyDetection: loaded model
+                AnomalyDetection or StreamingDetection: loaded model
         """
 
         with open(path, "rb") as f:
-            model = pickle.load(f)
+            metadata = dill.load(f)
 
-        self = model
-        self.cluster_PFSA_files = []
-        self.generate_PFSA_pngs()
+        model = metadata["modeltype"](**metadata["user_params"])
+        if model.quantize:
+            model.quantizer = Quantizer(n_quantizations=1, eps=-1)
+            model.quantizer.parameters = metadata["fitted_params"]["quantizer_parameters"]
+            model.quantizer._feature_order = metadata["fitted_params"]["quantizer_feature_order"]
+        model.cluster_labels = metadata["fitted_params"]["cluster_labels"]
+        model.cluster_PFSA_info = metadata["fitted_params"]["cluster_PFSA_info"]
+        model.PFSA_llk_means = metadata["fitted_params"]["PFSA_llk_means"]
+        model.PFSA_llk_stds = metadata["fitted_params"]["PFSA_llk_stds"]
 
         # write PFSA files
-        for i in range(self.n_clusters):
-            self.cluster_PFSA_files.append(RANDOM_NAME(path=self.temp_dir))
-            with open(self.cluster_PFSA_files[i], "w") as f:
-                f.write(f"{self.__format_PFSA_info(self.cluster_PFSA_info[i])}\n")
+        model.cluster_PFSA_files = []
+        for i in range(model.n_clusters):
+            model.cluster_PFSA_files.append(RANDOM_NAME(path=model.temp_dir))
+            with open(model.cluster_PFSA_files[i], "w") as f:
+                f.write(f"{model.__format_PFSA_info(model.cluster_PFSA_info[i])}\n")
+        model.generate_PFSA_pngs()
 
-        self.generate_PFSA_pngs()
+        model.fitted = True
 
-        return self
+        return model
 
 
     def print_PFSAs(self):
@@ -207,11 +224,10 @@ class AnomalyDetection:
             Returns:
                 list[str]: list of file paths to png files
         """
-        if self.cluster_PFSA_pngs is None:
-            self.cluster_PFSA_pngs = []
-            for i in range(self.n_clusters):
-                self.cluster_PFSA_pngs.append(RANDOM_NAME(path=self.temp_dir))
-                DrawPFSA(pfsafile=self.cluster_PFSA_files, graphpref=self.cluster_PFSA_pngs[i]).run()
+        self.cluster_PFSA_pngs = []
+        for i in range(self.n_clusters):
+            self.cluster_PFSA_pngs.append(RANDOM_NAME(path=self.temp_dir))
+            DrawPFSA(pfsafile=self.cluster_PFSA_files[i], graphpref=self.cluster_PFSA_pngs[i]).run()
 
         return self.cluster_PFSA_pngs
 
@@ -224,7 +240,7 @@ class AnomalyDetection:
         """
 
         if not self.quantize or self.quantize_type is None:
-            return X.copy(deep=False)
+            return X.copy(deep=False).astype(np.int8)
 
         if self.verbose:
             print("Quantizing...")
@@ -242,7 +258,11 @@ class AnomalyDetection:
             if not self.fitted:
                 self.quantizer = Quantizer(n_quantizations=1, epsilon=-1)
                 self.quantizer.fit(X)
-            return pd.concat([quantized for quantized in self.quantizer.transform(X)], axis=1)
+            return pd.concat(
+                [quantized for quantized in self.quantizer.transform(X)],
+                axis=1,
+                copy=False
+            ).astype(np.int8)
         else:
             raise ValueError(f"Unknown quantize type: {self.quantize_type}. Choose \"simple\" or \"complex\".")
 
@@ -257,8 +277,9 @@ class AnomalyDetection:
             self.dist_matrix = self.quantized_data
         else:
             self.dist_matrix = pd.DataFrame(
-                Lsmash(data=self.quantized_data, data_type="symbolic", sae=False).run()
-            ).round(8)
+                Lsmash(data=self.quantized_data, data_type="symbolic", sae=False).run(),
+                dtype=np.float32
+            )
 
 
     def __calculate_cluster_labels(self):
@@ -278,10 +299,10 @@ class AnomalyDetection:
         n_clusters = len(set(cluster_labels)) - (1 if -1 in cluster_labels else 0)
 
         # reassign clusters such that cluster 0 is the most common label, 1 is the second most common, etc.
-        cluster_counts = np.zeros(n_clusters, dtype=np.int64)
+        cluster_counts = np.zeros(n_clusters, dtype=np.int32)
         for cluster in cluster_labels:
             cluster_counts[cluster] += 1
-        cluster_rank = np.full(n_clusters, n_clusters - 1, dtype=np.int64) - np.argsort(np.argsort(cluster_counts))
+        cluster_rank = np.full(n_clusters, n_clusters - 1, dtype=np.int32) - np.argsort(np.argsort(cluster_counts))
         cluster_labels = [cluster_rank[cluster] for cluster in cluster_labels]
 
         self.n_clusters = n_clusters
@@ -334,13 +355,13 @@ class AnomalyDetection:
         if self.verbose:
             print("Calculating cluster PFSA means and stds...")
 
-        PFSA_llk_means = []
-        PFSA_llk_stds = []
+        PFSA_llk_means = np.empty(shape=self.n_clusters)
+        PFSA_llk_stds = np.empty(shape=self.n_clusters)
         for i in range(self.n_clusters):
             cluster_data = self.quantized_data[self.quantized_data["cluster"] == i].drop(columns=["cluster"], axis=1)
-            llks = np.asarray(Llk(data=cluster_data, pfsafile=self.cluster_PFSA_files[i]).run())
-            PFSA_llk_means.append(np.mean(llks))
-            PFSA_llk_stds.append(np.std(llks))
+            llks = np.asarray(Llk(data=cluster_data, pfsafile=self.cluster_PFSA_files[i]).run(), dtype=np.float32)
+            PFSA_llk_means[i] = np.mean(llks)
+            PFSA_llk_stds[i] = np.std(llks, ddof=1)
 
         self.PFSA_llk_means, self.PFSA_llk_stds = PFSA_llk_means, PFSA_llk_stds
 
@@ -357,17 +378,21 @@ class AnomalyDetection:
             curr_seqs = pd.DataFrame(Prun(pfsafile=pfsa, data_len=10000, num_repeats=seqs_per_pfsa).run())
             seqs = pd.concat([seqs, curr_seqs], axis=0)
 
-        self.predict(seqs)
-        mismatches = pd.concat([pd.DataFrame(self.closest_match), pd.DataFrame(self.cluster_labels)], axis=1)
 
+        cluster_llks = []
+        labels = range(self.n_clusters)
         for i, pfsa in enumerate(self.cluster_PFSA_files):
-            curr_llks = np.array(Llk(data=seqs, pfsafile=pfsa).run())
+            curr_llks = np.asarray(Llk(data=seqs, pfsafile=pfsa).run())
             diffs = curr_llks - self.PFSA_llk_means[i]
+            cluster_llks.append(curr_llks)
             # print(i, curr_llks)
             # print(i, diffs)
             # print(f"mean: {self.PFSA_llk_means[i]}, std: {self.PFSA_llk_stds[i]}")
             # print("\n")
 
+        self.closest_match = np.argmin(cluster_llks, axis=0)
+        self.mismatches = pd.concat([pd.DataFrame(self.closest_match), pd.DataFrame(labels)], axis=1)
+        self.cluster_llks = cluster_llks
 
     def __format_PFSA_info(self, PFSA_info, indent_level=0):
         """ Format the PFSA information for output
@@ -431,17 +456,20 @@ class StreamingDetection(AnomalyDetection):
         self.window_size = window_size
         self.window_overlap = window_overlap
 
+
     def fit(self, X, y=None):
         X_split_streams = self.__split_streams(X)
         super().fit(X_split_streams)
 
+
     def predict(self, X=None):
         if X is None:
-            return super().predict(X)
+            return super().predict()
         else:
             X_split_streams = self.__split_streams(X)
             return super().predict(X_split_streams)
 
+
     def __split_streams(self, X):
         """ Split stream data into individual streams of length self.window_size
             that overlap by self.overlap
@@ -460,4 +488,3 @@ class StreamingDetection(AnomalyDetection):
             [X[beg(i):end(i)].reset_index(drop=True) for i in range(size)],
             axis=1
         ).T
-
